/* Includes ------------------------------------------------------------------*/
#include "simcom_http.h"

#if SIMCOM > 0
#if CELLUAR_HTTP_ENABLE > 0
/**
  * @brief  SIMCom http service initialize
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_InitializeHTTPService(void)
{
	Celluar_ErrorStatus eResult;
	char cCommand[16];
	
	Celluar_Take();
	sprintf(cCommand, "AT+HTTPINIT\r");
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "OK", NULL, NULL, NULL, CELLUAR_WAITTIME_500MS);
	Celluar_Give();
	return eResult;
}

/**
  * @brief  SIMCom set http parameter value
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_SetHTTPParametersValue(char *pcParameter, char *pcValue, unsigned long ulValue)
{
	Celluar_ErrorStatus eResult;
	char cCommand[200];
	
	if (pcParameter == NULL)
		return Celluar_Error;
	
	Celluar_Take();
	if (pcValue == NULL)
		sprintf(cCommand, "AT+HTTPPARA=\"%s\",%lu\r", pcParameter, ulValue);
	else
		sprintf(cCommand, "AT+HTTPPARA=\"%s\",\"%s\"\r", pcParameter, pcValue);
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "OK", NULL, NULL, NULL, CELLUAR_WAITTIME_5S);
	Celluar_Give();
	return eResult;
}

/**
  * @brief  SIMCom http action
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_HTTPAction(HTTPMethod eMethod, unsigned long *pulResponseLen, unsigned long *pulErr)
{
	Celluar_ErrorStatus eResult;
	char cCommand[16], cResponse[48];
	
	Celluar_Take();
	sprintf(cCommand, "AT+HTTPACTION=%d\r", eMethod);
	if ((eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "+HTTPACTION:", cResponse, sizeof(cResponse), NULL, CELLUAR_WAITTIME_2M)) == Celluar_OK)
	{
		char *pcSplitedData[3] = {NULL, NULL, NULL};
		
		Split(pcSplitedData, 3, cResponse, ",");
		if (pcSplitedData[1] != NULL && pcSplitedData[2] != NULL)
		{
			unsigned short usStatus;
			
			usStatus = atol(pcSplitedData[1]);
			if (pulErr != NULL)
				*pulErr = usStatus;
			if (usStatus != 200)
			{
				eResult = Celluar_Error;
			}
			else
			{
				if (pulResponseLen != NULL)
				{
					*pulResponseLen = atol(pcSplitedData[2]);
				}
			}
		}
	}
	Celluar_Give();
	return eResult;
}

/**
  * @brief  SIMCom http input post data
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_InputHTTPData(unsigned char *pucData, unsigned long ulLen, unsigned long ulTimeout)
{
	Celluar_ErrorStatus eResult;
	char cCommand[24];
	
	if (pucData == NULL)
		return Celluar_Error;
	
	Celluar_Take();
	sprintf(cCommand, "AT+HTTPDATA=%lu,%lu\r", ulLen, ulTimeout);
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "DOWNLOAD", NULL, NULL, NULL, CELLUAR_WAITTIME_30S);
	if (eResult == Celluar_OK)
	{
		eResult = Celluar_SendCommandAndWaitForResponse((char *)pucData, ulLen, "OK", NULL, NULL, NULL, ulTimeout);
	}
	Celluar_Give();
	return eResult;
}

/**
  * @brief  SIMCom read http server response
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_ReadHTTPServerResponse(unsigned long ulStartAddress, unsigned long ulSize, unsigned char *pucResponse, unsigned long ulBufferLen, unsigned long *pulRespLen)
{
	Celluar_ErrorStatus eResult;
	char cCommand[24];
	unsigned long ulRecvBufferLen;
	
	Celluar_Take();
	if (ulSize == 0)
		sprintf(cCommand, "AT+HTTPREAD\r");
	else
		sprintf(cCommand, "AT+HTTPREAD=%lu,%lu\r", ulStartAddress, ulSize);
	if ((eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "+HTTPREAD:", (char *)pucResponse, ulBufferLen, &ulRecvBufferLen, CELLUAR_WAITTIME_10S)) == Celluar_OK)
	{
		char *pcStart = strstr((char *)pucResponse, "+HTTPREAD:");
		char *pcData = strstr(pcStart, "\r\n");		
		
		if (pcStart != NULL && pcData != NULL && pucResponse != NULL)
		{
			unsigned long ulLen;
			unsigned long ulRecvDataLen;
			
			//Point to data area
			pcStart += strlen("+HTTPREAD:");
			pcData += strlen("\r\n");
			ulLen = atol(pcStart);
			if (pulRespLen != NULL)
				*pulRespLen = ulLen;
			
			ulRecvDataLen = ulRecvBufferLen - (pcData - (char *)pucResponse);
			if (ulRecvDataLen >= ulLen)
			{
				memmove(pucResponse, pcData, ulLen);
			}
			else
			{
				memset(pucResponse, 0, ulBufferLen);
				eResult = Celluar_Error;
			}
		}
	}
	Celluar_Give();
	return eResult;
}

/**
  * @brief  SIMCom terminate http service
  * @retval Celluar_ErrorStatus
  */
Celluar_ErrorStatus SIMCom_TerminateHTTPService(void)
{
	Celluar_ErrorStatus eResult;
	char cCommand[16];
	
	Celluar_Take();
	sprintf(cCommand, "AT+HTTPTERM\r");
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "OK", NULL, NULL, NULL, CELLUAR_WAITTIME_500MS);
	Celluar_Give();
	return eResult;
}
#endif
#endif
