/* Includes ------------------------------------------------------------------*/
#include "celluar_comm.h"
#include "celluar_usb.h"
#include "celluar_gpio.h"
#include "simcom.h"
#include "quectel.h"
#include "esp_log.h"

/* Private macros ------------------------------------------------------------*/
#define CELLUAR_ERROR_MESSAGE			"ERROR"
#define CELLUAR_CLOSED_MESSAGE			"CLOSED"
#define CELLUAR_PDP_DEACT_MESSAGE		"+PDP: DEACT"
#define CELLUAR_ERROR_MESSAGE_2			"ERROR"
#define CELLUAR_CLOSED_MESSAGE_2		"+QIURC: \"closed\""
#define CELLUAR_PDP_DEACT_MESSAGE_2		"+QIURC: \"pdpdeact\""
#define CELLUAR_MAX_SEM_COUNT			3
#define CELLUAR_INITIAL_SEM_COUNT		0

/* Variables -----------------------------------------------------------------*/
Celluar g_xCelluar;
#if CELLUAR_TCPIP_ENABLE > 0
char *g_pcAPNList[CELLUAR_APN_COUNT] = 
{
	"cmnet",
	"3gnet",
	"globaldata",
	"mb",
	"unim2m.njm2mapn",
	"cuiot"
};
#endif

#if CELLUAR_NTP_ENABLE > 0
char *g_pcNTPServerList[CELLUAR_NTP_SERVER_COUNT] = 
{
	"202.108.6.95",
	"58.220.207.226",
	"47.92.108.218",
	"202.112.29.82",
	"120.25.108.11",
	"182.92.12.11",
	"115.28.122.198",
	"202.112.31.197",
	"202.112.29.82",
	"202.118.1.130"
};
#endif

void 				(*pFuncCelluar_Poll)(void);	
void 				(*pFuncCelluar_BasicSettings)(void);
#if CELLUAR_TCPIP_ENABLE > 0
void 				(*pFuncCelluar_TCPIPInit)(unsigned long *);
Celluar_ErrorStatus (*pFuncCelluar_Connect)(unsigned long);
Celluar_ErrorStatus (*pFuncCelluar_Send)(unsigned long, unsigned char *, unsigned long);
Celluar_ErrorStatus (*pFuncCelluar_GetSocketStatus)(unsigned long);
Celluar_ErrorStatus (*pFuncCelluar_Disconnect)(unsigned long);
#if CELLUAR_USE_SSL > 0
Celluar_ErrorStatus (*pFuncCelluar_SetSSL)(bool);
#endif
#endif
#if CELLUAR_TCPIP_ENABLE > 0 || CELLUAR_SMS_ENABLE > 0
Celluar_ErrorStatus (*pFuncCelluar_GetCCID)(char *, unsigned long);
Celluar_ErrorStatus (*pFuncCelluar_GetSignal)(unsigned char *, unsigned char *);
#endif
#if CELLUAR_GNSS_ENABLE > 0
Celluar_ErrorStatus (*pFuncCelluar_GetGPS)(double *, double *);
Celluar_ErrorStatus (*pFuncCelluar_GetCellArrayInfo)(unsigned long *, unsigned long *, unsigned long *, unsigned long *, unsigned long *, unsigned long *);
#endif
#if CELLUAR_NTP_ENABLE > 0
bool (*pFuncCelluar_SyncNetworkTime)(Calendar *);
#endif
#if CELLUAR_HTTP_ENABLE > 0
bool (*pFuncCelluar_HTTPPost)(char *, unsigned char *, unsigned long , unsigned char *, unsigned long, unsigned long *, unsigned long *);
bool (*pFuncCelluar_HTTPGet)(char *, unsigned char *, unsigned long, unsigned long *, unsigned long *);
Celluar_ErrorStatus (*pFuncCelluar_ReadHTTPServerResponse)(unsigned long, unsigned long, unsigned char *, unsigned long, unsigned long *);
Celluar_ErrorStatus (*pFuncCelluar_TerminateHTTPService)(void);
#endif

/* Private functions ---------------------------------------------------------*/
static void Celluar_ConfigureSem(void);
static void Celluar_ResetParmater(void);
static void Celluar_RxneHandler(unsigned char data);
static void Celluar_IdleHandler();
static void Celluar_DelayWhenPowerOn(void);
static void Celluar_DelayWhenPowerOff(void);


/**
  * @brief  Celluar Init
  * @param  None
  * @retval None
  */
void Celluar_Init()
{
	Celluar_ConfigureSem();
	Celluar_ConfigureGPIO();
	Celluar_ConfigureUSB(Celluar_IdleHandler, Celluar_RxneHandler);
}

/**
  * @brief  Celluar restart
  * @param  None
  * @retval None
  */
void Celluar_Restart()
{	
	Celluar_ResetParmater();
	if (Celluar_IsPowerOnWithAT())
	{
		//Off
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOff();
		//On
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOn();
	}
	else
	{
		//On
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOn();
	}
	
	if (Celluar_IsPowerOnWithAT() == true)
	{
		g_xCelluar.bIsPowerOn = true;
	}
}

/**
  * @brief  Celluar check if power on with AT command
  * @param  None
  * @retval None
  */
bool Celluar_IsPowerOnWithAT()
{
    Celluar_ErrorStatus eResult;
    
    Celluar_Take();
	eResult = Celluar_SendCommandAndWaitForResponse("AT\r", strlen("AT\r"), "OK", NULL, 0, NULL, CELLUAR_WAITTIME_500MS);
    Celluar_Give();
    
	return eResult != Celluar_Timeout ? true : false;
}

bool Celluar_IsSoftReboot(void)
{
    Celluar_ErrorStatus eResult;
    
    Celluar_Take();
	eResult = Celluar_SendCommandAndWaitForResponse("AT+CFUN=1,1\r", strlen("AT+CFUN=1,1\r"), "OK", NULL, 0, NULL, CELLUAR_WAITTIME_500MS);
    Celluar_Give();
    // printf("SoftReboot %d\r\n", eResult);
	return eResult != Celluar_Timeout ? true : false;
}


bool Celluar_PowerOnCheckAT()
{
	bool eState = false;
	if (Celluar_IsPowerOnWithAT())
	{
		eState = true;
	}
	else
	{
		if (Celluar_IsPowerOnWithAT())
		{
			eState = true;
		}
		else
		{
			Celluar_TriggerPowerKey();
			Celluar_DelayWhenPowerOn();
			if (Celluar_IsPowerOnWithAT() == true)
			{
				eState = true;
				Celluar_ResetParmater();
			}
		}
	}
	
	g_xCelluar.bIsPowerOn = (eState == true ? true : false);
	return eState;
}

bool Celluar_PowerOffCheckAT()
{
	bool eState = false;
	
	if (!Celluar_IsPowerOnWithAT())
	{
		eState = true;
	}
	else
	{
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOff();
		eState = (bool)!Celluar_IsPowerOnWithAT();
	}
	
	if (eState == true)
		Celluar_ResetParmater();
	
	return eState;
}

/**
  * @brief  Celluar power on 
  * @param  None
  * @retval is power on ? (true/false)
  */
bool Celluar_PowerOnCheckGPIO()
{
	if (Celluar_IsPowerOnWithGPIO())
	{
		return true;
	}
	else
	{
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOn();
		return Celluar_IsPowerOnWithGPIO();
	}
}

/**
  * @brief  Celluar power off 
  * @param  None
  * @retval is power off ? (true/false)
  */
bool Celluar_PowerOffCheckGPIO()
{
	if (!Celluar_IsPowerOnWithGPIO())
	{
		return true;
	}
	else
	{
		Celluar_TriggerPowerKey();
		Celluar_DelayWhenPowerOff();
		return (bool)!Celluar_IsPowerOnWithGPIO();
	}
}

/**
  * @brief  Celluar Sem Lock
  * @param  None
  * @retval None
  */
void Celluar_Take()
{
	xSemaphoreTake(g_xCelluar.xCommandSemHandle, portMAX_DELAY);
}

/**
  * @brief  Celluar Sem Unlock
  * @param  None
  * @retval None
  */
void Celluar_Give()
{
	xSemaphoreGive(g_xCelluar.xCommandSemHandle);
}

/**
  * @brief  Celluar's status is OK
  * @param  None
  * @retval true / false
  */
bool Celluar_IsNetStatusOK()
{
	return (bool)(g_xCelluar.bIsPowerOn && g_xCelluar.bHasSIMCard && g_xCelluar.bHasNetwork);
}

bool Celluar_HasSIMCard()
{
	return g_xCelluar.bHasSIMCard;
}

bool Celluar_IsNetAttached()
{
	return g_xCelluar.bHasNetwork;
}

/**
  * @brief  Celluar power on status
  * @param  None
  * @retval true / false
  */
bool Celluar_IsPowerOn()
{
	return g_xCelluar.bIsPowerOn;
}

/**
  * @brief  Celluar AT Command Communication
  * @param  pcCommand: Command string pointer
  * @param  ulCommandLen: Length of command
  * @param  pcDesiredResponse  Desired response pointer
  * @param  pcResponseData:  Response message
  * @param  ulResponseBufferSize:  Response buffer size 
  * @param  ulWaitTime  the time you wait for
  * @retval true: Success / false: Timeout or has not received desired response
  */
Celluar_ErrorStatus Celluar_SendCommandAndWaitForResponse(char *pcCommand, unsigned long ulCommandLen, char *pcDesiredResponse, char *pcResponseData, unsigned long ulResponseBufferSize, unsigned long *pulResponseLen, unsigned long ulWaitTime)
{
	BaseType_t			xResult;
	Celluar_ErrorStatus	eStatus;
	//Cal time
	TickType_t 			xEnterInTime = 0;
	TickType_t			xElapseTime = 0;
	TickType_t			xRestTime = 0;
	
	//Send Command
	 Celluar_Send((unsigned char *)pcCommand, ulCommandLen);
	
	//needn't response
	if (pcDesiredResponse == NULL)
		return Celluar_OK;
	
	//Wait Response
	g_xCelluar.bWaiting = true;
	//store current time
	xRestTime 		= (TickType_t)ulWaitTime;
	xEnterInTime 	= xTaskGetTickCount();
	while (true)
	{	
		//Calculate elapse time
		xElapseTime = xTaskGetTickCount() - xEnterInTime;
		//Calculate rest time
		xRestTime = xElapseTime > (TickType_t)ulWaitTime ? 0 : (TickType_t)ulWaitTime - xElapseTime;
		//Wating
		if (xRestTime)
			xResult = xSemaphoreTake(g_xCelluar.xResponseSemHandle, xRestTime);
		else
			xResult = pdFALSE;

		if (xResult == pdTRUE)
		{
			Celluar_Buffer *pxBuffer = &g_xCelluar.xReceiveQueue.xReceiveBuffer[g_xCelluar.xReceiveQueue.ulFront];
			char *pcPointer = NULL;
			
			if ((pcPointer = memstr(pxBuffer->ucBuffer, pcDesiredResponse, pxBuffer->ulBufferLen)) != NULL)
			{
				if (pcResponseData != NULL && ulResponseBufferSize > 0)
				{
					unsigned long ulNeedCopySize = 0;
					//copy pre (....... \r\nOK\r\n)
					if (memcmp(pcDesiredResponse, "OK", sizeof("OK")) == 0)
					{
						unsigned char *pcHeader = pxBuffer->ucBuffer;
						while (pcHeader + strlen((char *)pcHeader) < (unsigned char *)pcPointer)
						{
							unsigned long ulTmpLen = strlen((char *)pcHeader);
							pcHeader += (ulTmpLen + 1);
						}
						ulNeedCopySize = (unsigned char *)pcPointer - pcHeader;
						if (ulResponseBufferSize > ulNeedCopySize)
						{
							memcpy(pcResponseData, (char *)pcHeader, ulNeedCopySize);
							pcResponseData[ulNeedCopySize] = '\0';
						}
						else
						{
							memcpy(pcResponseData, pcPointer - (ulResponseBufferSize - 1), ulResponseBufferSize - 1);
							pcResponseData[ulResponseBufferSize - 1] = '\0';
						}
					}
					else	//copy last(+CPIN: ......)
					{	
						ulNeedCopySize = pxBuffer->ucBuffer + pxBuffer->ulBufferLen - (unsigned char *)pcPointer;
						if (ulResponseBufferSize > ulNeedCopySize)
						{
							memcpy(pcResponseData, pcPointer, ulNeedCopySize);
							pcResponseData[ulNeedCopySize] = '\0';
						}
						else
						{
							memcpy(pcResponseData, pcPointer, ulResponseBufferSize - 1);
							pcResponseData[ulResponseBufferSize - 1] = '\0';
						}
						
						if (pulResponseLen != NULL)
							*pulResponseLen = ulResponseBufferSize > ulNeedCopySize ? ulNeedCopySize : ulResponseBufferSize - 1;
					}
				}
				
				eStatus = Celluar_OK;
				Celluar_QueuePop(&g_xCelluar.xReceiveQueue);
				break;
			}
			else if ((g_xCelluar.eModule == Celluar_SIMCom && strstr(pcDesiredResponse, "SEND") && (memstr(pxBuffer->ucBuffer, CELLUAR_CLOSED_MESSAGE, pxBuffer->ulBufferLen) || memstr(pxBuffer->ucBuffer, CELLUAR_PDP_DEACT_MESSAGE, pxBuffer->ulBufferLen))) ||
					 (g_xCelluar.eModule == Celluar_Quectel && strstr(pcDesiredResponse, "SEND") && (memstr(pxBuffer->ucBuffer, CELLUAR_CLOSED_MESSAGE_2, pxBuffer->ulBufferLen) || memstr(pxBuffer->ucBuffer, CELLUAR_PDP_DEACT_MESSAGE_2, pxBuffer->ulBufferLen))))
			{
				eStatus = Celluar_Error;
				Celluar_QueuePop(&g_xCelluar.xReceiveQueue);
				break;
			}
			else if ((g_xCelluar.eModule == Celluar_SIMCom && strstr(pcDesiredResponse, "CONNECT") && (memstr(pxBuffer->ucBuffer, CELLUAR_CLOSED_MESSAGE, pxBuffer->ulBufferLen) || memstr(pxBuffer->ucBuffer, CELLUAR_PDP_DEACT_MESSAGE, pxBuffer->ulBufferLen))) ||
					 (g_xCelluar.eModule == Celluar_Quectel && strstr(pcDesiredResponse, "+QIOPEN:") && (memstr(pxBuffer->ucBuffer, CELLUAR_CLOSED_MESSAGE_2, pxBuffer->ulBufferLen) || memstr(pxBuffer->ucBuffer, CELLUAR_PDP_DEACT_MESSAGE_2, pxBuffer->ulBufferLen))))
			{
				eStatus = Celluar_Error;
				Celluar_QueuePop(&g_xCelluar.xReceiveQueue);
				break;
			}
			else if (memstr(pxBuffer->ucBuffer, CELLUAR_ERROR_MESSAGE, pxBuffer->ulBufferLen))
			{
				eStatus = Celluar_Error;
				Celluar_QueuePop(&g_xCelluar.xReceiveQueue);
				break;
			}
			else
			{
				Celluar_QueuePop(&g_xCelluar.xReceiveQueue);
				continue;
			}
		}
		else
		{
			eStatus = Celluar_Timeout;
			break;
		}
	}
	
	g_xCelluar.bWaiting = false;
	return eStatus;
}

/**
  * @brief  Celluar set specific module's functions
  * @param  None
  * @retval None
  */
void Celluar_SetFunc(Celluar_Module xModule)
{
	switch(xModule)
	{
	#if SIMCOM > 0
		case Celluar_SIMCom:
		#if CELLUAR_TCPIP_ENABLE > 0
			#if CELLUAR_MULTIIPCONNECTION_ENABLE > 0
				g_ulMaxSocketID = 6 > CELLUAR_SOCKET_MAX_COUNT ? CELLUAR_SOCKET_MAX_COUNT : 6;
			#else
				//Single
				g_ulMaxSocketID = 1;
			#endif
		#endif
			pFuncCelluar_Poll 				= SIMCom_Poll;
			pFuncCelluar_BasicSettings		= SIMCom_BasicSettings;
		#if CELLUAR_TCPIP_ENABLE > 0
			pFuncCelluar_TCPIPInit			= SIMCom_TCPIPInit;
			pFuncCelluar_Connect			= SIMCom_Connect;
			pFuncCelluar_Send				= SIMCom_Send;
			pFuncCelluar_GetSocketStatus	= SIMCom_GetSocketStatus;
			pFuncCelluar_Disconnect			= SIMCom_Disconnect;
		#if CELLUAR_USE_SSL > 0
			pFuncCelluar_SetSSL    			= SIMCom_SetSSL;
		#endif
		#endif
		#if CELLUAR_TCPIP_ENABLE > 0 || CELLUAR_SMS_ENABLE > 0
			pFuncCelluar_GetSignal			= SIMCom_GetSignal;
			pFuncCelluar_GetCCID			= SIMCom_GetCCID;
		#endif
		#if CELLUAR_GNSS_ENABLE > 0
			pFuncCelluar_GetGPS				= SIMCom_GetGNSSNavigationInfomation;
			pFuncCelluar_GetCellArrayInfo	= SIMCom_GetCellArrayInfo;
		#endif
		#if CELLUAR_NTP_ENABLE > 0
			pFuncCelluar_SyncNetworkTime 	= SIMCom_SyncNetworkTime;
		#endif
		#if CELLUAR_HTTP_ENABLE > 0
			pFuncCelluar_HTTPPost			= SIMCom_HTTPPost;
			pFuncCelluar_HTTPGet			= SIMCom_HTTPGet;
			pFuncCelluar_ReadHTTPServerResponse = SIMCom_ReadHTTPServerResponse;
			pFuncCelluar_TerminateHTTPService = SIMCom_TerminateHTTPService;
		#endif
			xSemaphoreGive(g_xCelluar.xReceivedSemHandle);
			break;
	#endif
	#if QUECTEL > 0
		case Celluar_Quectel:
		#if CELLUAR_TCPIP_ENABLE > 0
			#if CELLUAR_MULTIIPCONNECTION_ENABLE > 0
				g_ulMaxSocketID = 6 > CELLUAR_SOCKET_MAX_COUNT ? CELLUAR_SOCKET_MAX_COUNT : 6;
			#else
				//Single socket
				g_ulMaxSocketID = 1;
			#endif
		#endif
			pFuncCelluar_Poll 				= Quectel_Poll;
			pFuncCelluar_BasicSettings		= Quectel_BasicSettings;
		#if CELLUAR_TCPIP_ENABLE > 0
			pFuncCelluar_TCPIPInit			= Quectel_TCPIPInit;
			pFuncCelluar_Connect			= Quectel_Connect;
			pFuncCelluar_Send				= Quectel_Send;
			pFuncCelluar_GetSocketStatus	= Quectel_GetSocketStatus;
			pFuncCelluar_Disconnect			= Quectel_Disconnect;
		#if CELLUAR_USE_SSL > 0
			pFuncCelluar_SetSSL    			= Quectel_SetSSL;
		#endif
		#endif
		#if CELLUAR_TCPIP_ENABLE > 0 || CELLUAR_SMS_ENABLE > 0
			pFuncCelluar_GetSignal			= Quectel_GetSignal;
			pFuncCelluar_GetCCID			= Quectel_GetCCID;
		#endif
		#if CELLUAR_GNSS_ENABLE > 0
			pFuncCelluar_GetGPS				= Quectel_GetGNSSNavigationInfomation;
			pFuncCelluar_GetCellArrayInfo	= Quectel_GetCellArrayInfo;
		#endif
		#if CELLUAR_NTP_ENABLE > 0
			pFuncCelluar_SyncNetworkTime 	= Quectel_SyncNetworkTime;
		#endif
		#if CELLUAR_HTTP_ENABLE > 0
			pFuncCelluar_HTTPPost			= Quectel_HTTPPost;
			pFuncCelluar_HTTPGet			= Quectel_HTTPGet;
			pFuncCelluar_ReadHTTPServerResponse = Quectel_ReadHTTPServerResponseFromRAM;
			pFuncCelluar_TerminateHTTPService = Quectel_TerminateHTTPService;
		#endif
			xSemaphoreGive(g_xCelluar.xReceivedSemHandle);
			break;
	#endif
		default:
			break;
	}
}

/**
  * @brief  Celluar sending ATI command
  * @param  None
  * @retval status
  */
Celluar_ErrorStatus	Celluar_GetModuleInfo(char *pcModuleInfo, unsigned long ulModuleBufferSize)
{
	Celluar_ErrorStatus eResult;
	char cCommand[8];
	
	Celluar_Take();
	sprintf(cCommand, "ATI\r");
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "OK", pcModuleInfo, ulModuleBufferSize, NULL, CELLUAR_WAITTIME_500MS);
	Celluar_Give();
	return eResult;
}

/**
  * @brief  Celluar sending ATEx command to set the echo mode
  * @param  None
  * @retval status
  */
Celluar_ErrorStatus Celluar_SetEchoMode(bool bStatus)
{
	Celluar_ErrorStatus eResult;
	char cCommand[8];
	
	Celluar_Take();
	sprintf(cCommand, "ATE%d\r", bStatus ? 1 : 0);
	eResult = Celluar_SendCommandAndWaitForResponse(cCommand, strlen(cCommand), "OK", NULL, 0, NULL, CELLUAR_WAITTIME_500MS);
	Celluar_Give();
	return eResult;
}

/**
  * @brief  Celluar sem init
  * @param  None
  * @retval None
  */
static void Celluar_ConfigureSem()
{
	if (g_xCelluar.xCommandSemHandle == NULL)
	{
		g_xCelluar.xCommandSemHandle = xSemaphoreCreateBinary();
		xSemaphoreGive(g_xCelluar.xCommandSemHandle);
	}
	
	if (g_xCelluar.xReceivedSemHandle == NULL)
	{
		g_xCelluar.xReceivedSemHandle = xSemaphoreCreateCounting(CELLUAR_MAX_SEM_COUNT, CELLUAR_INITIAL_SEM_COUNT);
	}
	
	if (g_xCelluar.xResponseSemHandle == NULL)
	{
		g_xCelluar.xResponseSemHandle = xSemaphoreCreateCounting(CELLUAR_MAX_SEM_COUNT, CELLUAR_INITIAL_SEM_COUNT);
	}
}

/**
  * @brief  Celluar reset parmater
  * @param  None
  * @retval None
  */
static void Celluar_ResetParmater()
{
	//reset parm
	g_xCelluar.bIsPowerOn = false;
	g_xCelluar.bHasSIMCard = false;
	g_xCelluar.bHasNetwork = false;
}

//IT Handler
static void Celluar_RxneHandler(unsigned char data)
{
	unsigned char ucReceived;	
	ucReceived = data;
	if (!Celluar_QueueIsFull(&g_xCelluar.xReceiveQueue))
	{
		Celluar_Buffer *pxBuffer = &g_xCelluar.xReceiveQueue.xReceiveBuffer[g_xCelluar.xReceiveQueue.ulRear];
		if (pxBuffer->ulBufferLen < CELLUAR_BUFFER_SIZE)
		{
			pxBuffer->ucBuffer[pxBuffer->ulBufferLen++] = ucReceived;
		}
	}
}

static void Celluar_IdleHandler()
{
	if (!Celluar_QueueIsFull(&g_xCelluar.xReceiveQueue))
	{
		//Received
		Celluar_QueuePush(&g_xCelluar.xReceiveQueue);
	}
	xSemaphoreGive(g_xCelluar.xReceivedSemHandle);
}

static void Celluar_DelayWhenPowerOn()
{
	if (g_xCelluar.eModule == Celluar_SIMCom)
	{
		//SIM800C serial port active after 3s when power on 
		vTaskDelay(3000);
	}
	else if (g_xCelluar.eModule == Celluar_Quectel)
	{
		//Quectel serial port active after 15s
		vTaskDelay(15000);
	}
	else
	{
		//Quectel serial port active after 15s
		vTaskDelay(15000);
	}
}

static void Celluar_DelayWhenPowerOff()
{
	if (g_xCelluar.eModule == Celluar_SIMCom)
	{
		//SIM800C serial port inactive after 2s when power on 
		vTaskDelay(3000);
	}
	else if (g_xCelluar.eModule == Celluar_Quectel)
	{
		//Quectel serial port inactive after 30s
		vTaskDelay(30000);
	}
	else
	{
		//Quectel serial port inactive after 30s
		vTaskDelay(30000);
	}
}


