/* Includes ------------------------------------------------------------------*/
#include <string.h>
#include "celluar_usb.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "freertos/event_groups.h"
#include "esp_log.h"
#include "unity.h"
#include "esp_usbh_cdc.h"
/* Private macros ------------------------------------------------------------*/
#define IN_RINGBUF_SIZE           (256 * 8)
#define OUT_RINGBUF_SIZE          (256 * 8)
#define READ_TASK_KILL_BIT        BIT1

#define TAG "cdc_usb"
static SemaphoreHandle_t gUSBSemaphore;
#define TEST_ITF_NUM  1

#define TEST_EP_IN   0x86
#define TEST_EP_OUT  0x0f

typedef struct
{
	void (*pFuncRxneHandler)(unsigned char);
  void (*pFuncIdleHandler)(void);
} USB_Handler;

static USB_Handler g_xUSBHandler;

size_t  data_len = 0;
uint8_t buf[IN_RINGBUF_SIZE];

static usb_ep_desc_t bulk_out_ep_desc[TEST_ITF_NUM] = {
  [0] = {
    .bLength = sizeof(usb_ep_desc_t),
    .bDescriptorType = USB_B_DESCRIPTOR_TYPE_ENDPOINT,
    .bEndpointAddress = TEST_EP_OUT,       //EP 1 OUT
    .bmAttributes = USB_BM_ATTRIBUTES_XFER_BULK,
    .wMaxPacketSize = 64,           //MPS of 64 bytes
    .bInterval = 0,
  }
};

static usb_ep_desc_t bulk_in_ep_desc[TEST_ITF_NUM] = {
  [0] = {
    .bLength = sizeof(usb_ep_desc_t),
    .bDescriptorType = USB_B_DESCRIPTOR_TYPE_ENDPOINT,
    .bEndpointAddress = TEST_EP_IN,       //EP 1 IN
    .bmAttributes = USB_BM_ATTRIBUTES_XFER_BULK,
    .wMaxPacketSize = 64,           //MPS of 64 bytes
    .bInterval = 0,  
  }
};

static usbh_cdc_config_t config = {
  .itf_num = TEST_ITF_NUM,
  .bulk_in_eps[0]     = &bulk_in_ep_desc[0],
  .bulk_out_eps[0]    = &bulk_out_ep_desc[0],
  .rx_buffer_sizes[0] = IN_RINGBUF_SIZE,
  .tx_buffer_sizes[0] = OUT_RINGBUF_SIZE,
};

static void usb_read_task(void *param)
{
    while (1) {
      xSemaphoreTake(gUSBSemaphore, portMAX_DELAY);
      if(usbh_cdc_get_itf_state(0)){
        usbh_cdc_itf_get_buffered_data_len(0, &data_len);
        if (data_len > 0) {
            usbh_cdc_itf_read_bytes(0, buf, data_len, 10);
            // ESP_LOGD(TAG, "rcv: %*s, len: %d \n", data_len, buf, data_len);
            // esp_log_buffer_char("usb r", buf, data_len);            
            for(size_t i = 0 ; i < data_len ; i++) {
              if (g_xUSBHandler.pFuncRxneHandler != NULL)
                g_xUSBHandler.pFuncRxneHandler(buf[i]);
            }
        }else{
          if (g_xUSBHandler.pFuncIdleHandler != NULL)
            g_xUSBHandler.pFuncIdleHandler();
        }
      }
      else{
        if(g_xUSBHandler.pFuncIdleHandler != NULL)
          g_xUSBHandler.pFuncIdleHandler();
      }
      xSemaphoreGive(gUSBSemaphore);
      vTaskDelay(5/ portTICK_PERIOD_MS);
    }
}

/**
  * @brief  Celluar USB Init
  * @param  None
  * @retval None
  */
void Celluar_ConfigureUSB(void (*pFuncIdleCallBack)(void), void (*pFuncRxneCallBack)(unsigned char))
{ 
    usbh_cdc_driver_install(&config);
    /* Waiting for USB device connected */  
    usbh_cdc_wait_connect(5000);  
    g_xUSBHandler.pFuncRxneHandler	= pFuncRxneCallBack;
    g_xUSBHandler.pFuncIdleHandler = pFuncIdleCallBack;
    gUSBSemaphore = xSemaphoreCreateBinary(); 
    xSemaphoreGive(gUSBSemaphore);
    xTaskCreate(usb_read_task, "usb_read", 3072, NULL, 5, NULL);
}

/**
  * @brief  Celluar UART Send
  * @param  pucData: Data pointer
  * @param  ulLen: Length of data
  * @retval None
  */
void Celluar_Send(unsigned char *pucData, unsigned long ulLen)
{
  xSemaphoreTake(gUSBSemaphore, portMAX_DELAY);
  // ESP_LOGD(TAG, "txc: %*s, len: %d \n",(int)ulLen, pucData, (int)ulLen);
  // esp_log_buffer_char("usb", pucData, ulLen);      
  usbh_cdc_itf_write_bytes(0, pucData, ulLen);
  xSemaphoreGive(gUSBSemaphore);
}

void Celluar_usb_deInit(void)
{
  usbh_cdc_driver_delete();
}

void Celluar_usb_Init(void)
{
  usbh_cdc_driver_install(&config);
  usbh_cdc_wait_connect(portMAX_DELAY);  
}