//*****************************************************************************
//! \file mqtt_interface.c
//! \brief Paho MQTT to WIZnet Chip interface implement file.
//! \details The process of porting an interface to use paho MQTT.
//! \version 1.0.0
//! \date 2016/12/06
//! \par  Revision history
//!       <2016/12/06> 1st Release
//!
//! \author Peter Bang & Justin Kim
//! \copyright
//!
//! Copyright (c)  2016, WIZnet Co., LTD.
//! All rights reserved.
//!
//! Redistribution and use in source and binary forms, with or without
//! modification, are permitted provided that the following conditions
//! are met:
//!
//!     * Redistributions of source code must retain the above copyright
//! notice, this list of conditions and the following disclaimer.
//!     * Redistributions in binary form must reproduce the above copyright
//! notice, this list of conditions and the following disclaimer in the
//! documentation and/or other materials provided with the distribution.
//!     * Neither the name of the <ORGANIZATION> nor the names of its
//! contributors may be used to endorse or promote products derived
//! from this software without specific prior written permission.
//!
//! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
//! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
//! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
//! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
//! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
//! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
//! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
//! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
//! THE POSSIBILITY OF SUCH DAMAGE.
//
//*****************************************************************************

#include "mqtt_interface.h"
#include "wit_config.h"
#include "mqtt_port.h"
#include <string.h>
#if CONFIG_W5100S_SPI_ETHERNET > 0
	#include "wizchip_conf.h"
	#include "socket.h"
#elif CONFIG_EC200S_USB_4G > 0
	#include "mqtts.h"
#endif
#include "esp_log.h"

int ThreadStart(Thread* thread, void (*fn)(void*), void* arg)
{
	int rc = 0;
	uint16_t usTaskStackSize = (configMINIMAL_STACK_SIZE * 5);
	UBaseType_t uxTaskPriority = uxTaskPriorityGet(NULL); /* set the priority as the same as the calling task*/

	rc = xTaskCreate(fn,	/* The function that implements the task. */
		"MQTTTask",			/* Just a text name for the task to aid debugging. */
		usTaskStackSize,	/* The stack size is defined in FreeRTOSIPConfig.h. */
		arg,				/* The task parameter, not used in this case. */
		uxTaskPriority,		/* The priority assigned to the task is defined in FreeRTOSConfig.h. */
		&thread->task);		/* The task handle is not used. */
		ESP_LOGD("rc","%d",rc);
	return rc;
}

void MutexInit(Mutex* mutex)
{
	mutex->sem = xSemaphoreCreateMutex();
}

int MutexLock(Mutex* mutex)
{
	return xSemaphoreTake(mutex->sem, portMAX_DELAY);
}

int MutexUnlock(Mutex* mutex)
{
	return xSemaphoreGive(mutex->sem);
}

void TimerInit(Timer* timer)
{
	timer->xTicksToWait = 0;
	memset(&timer->xTimeOut, '\0', sizeof(timer->xTimeOut));
}

char TimerIsExpired(Timer* timer)
{
	return xTaskCheckForTimeOut(&timer->xTimeOut, &timer->xTicksToWait) == pdTRUE;
}

void TimerCountdownMS(Timer* timer, unsigned int timeout_ms)
{
	timer->xTicksToWait = timeout_ms / portTICK_PERIOD_MS; /* convert milliseconds to ticks */
	vTaskSetTimeOutState(&timer->xTimeOut); /* Record the time at which this function was entered. */
}


void TimerCountdown(Timer* timer, unsigned int timeout) 
{
	TimerCountdownMS(timer, timeout * 1000);
}

int TimerLeftMS(Timer* timer) 
{
	xTaskCheckForTimeOut(&timer->xTimeOut, &timer->xTicksToWait); /* updates xTicksToWait to the number left */
	return (timer->xTicksToWait <= 0) ? 0 : (timer->xTicksToWait * portTICK_PERIOD_MS);
}

void NewNetwork(Network* n, int sn)
{ 
#if CONFIG_W5100S_SPI_ETHERNET > 0
	n->my_socket = sn;	
#endif
	n->mqttread = FreeRTOS_read;
	n->mqttwrite = FreeRTOS_write;
	n->disconnect = FreeRTOS_disconnect;	
}

int FreeRTOS_read(Network* n, unsigned char* buffer, int len, long timeout_ms)
{
	TickType_t xTicksToWait = timeout_ms / portTICK_PERIOD_MS; /* convert milliseconds to ticks */
	TimeOut_t xTimeOut;
	int recvLen = 0;
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		unsigned char tmp = 0;
	#endif
	vTaskSetTimeOutState(&xTimeOut); /* Record the time at which this function was entered. */
	do
	{
		int rc = 0;
		//Receive
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		// check_eth_ping(n->my_socket);
		check_phy(n->my_socket);
		tmp = getSn_SR(n->my_socket);
		if((tmp == SOCK_ESTABLISHED || tmp == SOCK_CLOSE_WAIT) && (getSn_RX_RSR(n->my_socket)>0))
		{
			rc = recv(n->my_socket, buffer + recvLen, len);		
			setMQTTStatusChangned(n->my_socket, true);	
		}
	#elif CONFIG_EC200S_USB_4G > 0
	   	rc = MQTT_RecvFromSeqQueue(n->my_socket, buffer + recvLen, len);
	#endif	
		if (rc > 0)
		{
			recvLen += rc;
		}
		else if (rc < 0 || (rc == 0 && len == 1))
		{
			recvLen = rc;
			break;
		}
		else
		{
			vTaskDelay(2);
		}
	} while (recvLen < len && xTaskCheckForTimeOut(&xTimeOut, &xTicksToWait) == pdFALSE);

	return recvLen;
}

int FreeRTOS_write(Network* n, unsigned char* buffer, int len, long timeout_ms)
{
	TickType_t xTicksToWait = timeout_ms / portTICK_PERIOD_MS; /* convert milliseconds to ticks */
	TimeOut_t xTimeOut;
	int sentLen = 0;
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		int sz = 0;
		int TXMAX = 0x700;
		unsigned char tmp = 0;
	#endif
	vTaskSetTimeOutState(&xTimeOut); /* Record the time at which this function was entered. */
	do
	{
		int rc = 0;
		//Send
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		check_phy(n->my_socket);
		for(int i = 0; i < len; i += TXMAX) 
		{
			tmp = getSn_SR(n->my_socket);
			if(tmp == SOCK_ESTABLISHED || tmp == SOCK_CLOSE_WAIT)
			{
				sz = (len - i) > TXMAX  ? TXMAX : (len - i);
				rc = send(n->my_socket, buffer + i, sz);	
				setMQTTStatusChangned(n->my_socket, true);	
				if(rc != sz)
				{
					setMQTTStatusChangned(n->my_socket, false);		
				}
    		}
			else
			{	
				rc = -1;
				printf("211 rc %d sz %d\r\n", rc, sz);
				setMQTTStatusChangned(n->my_socket, false);		
			}	
		}	
	#elif CONFIG_EC200S_USB_4G > 0
		if (MQTT_SendData(n->my_socket, buffer, len) == MQTT_TRUE)
		{
			rc = len;
		}
		else
		{
			rc = -1;
		}
	#endif
		
		if (rc > 0)
			sentLen += rc;
		else if (rc < 0)
		{
			sentLen = rc;
			break;
		}
	} while (sentLen < len && xTaskCheckForTimeOut(&xTimeOut, &xTicksToWait) == pdFALSE);

	return sentLen;
}

void FreeRTOS_disconnect(Network* n)
{
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		MQTT_SetSocketStatus(n->my_socket, false);
		disconnect(n->my_socket);	
	#elif CONFIG_EC200S_USB_4G > 0
		MQTT_SetSocketStatus(n->my_socket, false);
	#endif
}

int NetworkConnect(Network* n, char *addr, int port)
{
	#if CONFIG_W5100S_SPI_ETHERNET > 0
		uint16_t myport = 5000;
		// check_eth_ping(n->my_socket);
		check_phy(n->my_socket);
		if(my_socket(n->my_socket, Sn_MR_TCP, myport, 0) != n->my_socket)
			return SOCK_ERROR;
		if(addr == NULL)
			return SOCK_ERROR;
		if(connect(n->my_socket, (uint8_t *)addr, port) != SOCK_OK)
			return SOCK_ERROR;
		return SOCK_OK;
	#elif CONFIG_EC200S_USB_4G > 0
		return 0;	
	#endif
	return 0;
}
 


